/*****************************************************************************/
/* EyeClock - A clock application for X. The eyes follow the mouse.          */
/*                                                                           */
/* EyeClock Copyright (c) 1998 Sakai Hiroaki.                                */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* BSD Daemon Copyright 1988 by Marshall Kirk McKusick.                      */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* EyeClock is a simple clock application for X11. You can put an optional   */
/* picture on the face of the clock. And you can move the eyes of the        */
/* picture following the mouse.                                              */
/*                                                                           */
/* EyeClock is a free software. You may copy or distribute the original      */
/* EyeClock freely. But, you may not modify or distribute the original       */
/* EyeClock without permission of the author.                                */
/*                                                                           */
/* Statement on the Use of the BSD Daemon Figure:                            */
/*                                                                           */
/* The BSD Daemon is to be used in the context of BSD software.              */
/* So, if you are using BSD software (BSD/OS, FreeBSD, NetBSD, OpenBSD,      */
/* or BSD utilities) in your Intranet environment, then use of the           */
/* daemon is appropriate. If you are a Microsoft shop, then it is not        */
/* appropriate.                                                              */
/*                                                                           */
/* Individuals may use the daemon for their personal use within the          */
/* bounds of good taste (an example of bad taste was a picture of the        */
/* BSD daemon blowtorching a Solaris logo). When reasonably possible,        */
/* I would like the text ``BSD Daemon Copyright 1988 by Marshall Kirk        */
/* McKusick. All Rights Reserved.'' to be included.  This text need          */
/* not be etched into the figure or garishly displayed when using the        */
/* daemon as say an Icon in a Web window. A good example of how              */
/* to handle the due credit in a web page is to create a link from           */
/* the daemon picture to the following text:                                 */
/*                                                                           */
/*         BSD Daemon Copyright 1988 by Marshall Kirk McKusick.              */
/*         All Rights Reserved.                                              */
/*                                                                           */
/*         Permission to use the daemon may be obtained from:                */
/*                 Marshall Kirk McKusick                                    */
/*                 1614 Oxford St                                            */
/*                 Berkeley, CA 94709-1608                                   */
/*                 USA                                                       */
/*         or via email at mckusick@mckusick.com                             */
/*                                                                           */
/* If you are looking for daemon images, a pictorial history of the          */
/* daemon can be found at http://www.mckusick.com/beastie/index.html.        */
/*                                                                           */
/* If you want to mass produce the daemon on Tshirts, CDROM's, or other      */
/* products you need to request permission in advance. In general, I require */
/* that the daemon be used in an appropriate way. This means that it         */
/* has to be something related to BSD and not expropriated as a company      */
/* logo (though I do allow companies with BSD-based products such as         */
/* Walnut Creek CDROM or BSDI to use it).  I regret having to be so          */
/* legalistic about the daemon, but I almost lost the daemon to a            */
/* certain large company because I failed to show due diligence in           */
/* protecting it. So, I've taken due diligence seriously since then.         */
/*                                                                           */
/*         Marshall Kirk McKusick                                            */
/*****************************************************************************/

/*****************************************************************************/
/* إåΥ󥯥롼                                                      */
/*****************************************************************************/

#include"eyetype.h"

/*****************************************************************************/
/* פ                                                                */
/*****************************************************************************/

XPoint CalEyePoint(Eye eye, XPoint pointer, XPoint geo)
{
  XPoint a;
  a.x = (int)((0.00005 * eye.move.x * (pointer.x - (geo.x+eye.point.x))) -
	      eye.radius.x /2 + eye.center.x + eye.point.x);
  a.y = (int)((0.00005 * eye.move.y * (pointer.y - (geo.y+eye.point.y))) -
	      eye.radius.y /2 + eye.center.y + eye.point.y);
  return (a);
}

void MakeEyePixmap(EyeClockWindow * wp, Clock * cp, Eye * ep,
		   XPoint pointer, XPoint geo)
{
  XPoint a;

  XCopyArea(wp->display, wp->figure_pixmap, wp->clock_pixmap, wp->gc_copy_and,
	    0, 0, cp->width, cp->height, 0, 0);
  usleep(1000);
  XCopyArea(wp->display, wp->eye_pixmap, wp->work_pixmap, wp->gc_copy_copy,
	    0, 0, cp->width, cp->height, 0, 0);
  usleep(1000);
  while (ep != NULL) {
    a = CalEyePoint(*ep, pointer, geo);
    XFillArc(wp->display, wp->work_pixmap, ep->gc_eye_foreground,
	     a.x, a.y, ep->radius.x, ep->radius.y, 0, 360*64);
    ep = ep->next;
  }
  XCopyArea(wp->display, wp->figure_pixmap, wp->work_pixmap,
	    wp->gc_copy_andinv, 0, 0, cp->width, cp->height, 0, 0);
  usleep(1000);
  XCopyArea(wp->display, wp->work_pixmap, wp->clock_pixmap, wp->gc_copy_or,
	    0, 0, cp->width, cp->height, 0, 0);
  usleep(1000);
}

void MakeClockPixmap(EyeClockWindow * wp, Clock * cp, int hour, int minute)
{
  double rad_short, rad_long;
  XPoint s, se, le;

  rad_short = ((double)30 * hour + (double)minute/2 - 90)*DEGTORAD;
  rad_long = ((double)6 * minute - 90)*DEGTORAD;
  s.x = cp->clock_center.x;
  s.y = cp->clock_center.y;
  se.x = s.x + (int)(cp->short_length * cos(rad_short));
  se.y = s.y + (int)(cp->short_length * sin(rad_short));
  le.x = s.x + (int)( cp->long_length * cos(rad_long ));
  le.y = s.y + (int)( cp->long_length * sin(rad_long ));
  XDrawLine(wp->display,wp->clock_pixmap,wp->gc_short_back,s.x,s.y,se.x,se.y);
  XDrawLine(wp->display,wp->clock_pixmap,wp->gc_short,     s.x,s.y,se.x,se.y);
  XDrawLine(wp->display,wp->clock_pixmap,wp->gc_long_back, s.x,s.y,le.x,le.y);
  XDrawLine(wp->display,wp->clock_pixmap,wp->gc_long,      s.x,s.y,le.x,le.y);
}

void DrawClock(EyeClockWindow * wp, Clock * cp, Eye * ep,
	       int hour, int minute, XPoint pointer, XPoint geo)
{
  XCopyArea(wp->display, wp->back_pixmap, wp->clock_pixmap, wp->gc_copy_copy,
            0, 0, cp->width, cp->height, 0, 0);
  usleep(1000);
  if (cp->eye_flag == ON) MakeEyePixmap(wp, cp, ep, pointer, geo);
  if (cp->clock_flag == ON) MakeClockPixmap(wp, cp, hour, minute);
  XCopyArea(wp->display, wp->clock_pixmap, wp->clock_window, wp->gc_copy_copy,
	    0, 0, cp->width, cp->height, 0, 0);
  usleep(1000);
}

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

void SetParameter(EyeClockWindow * wp, Clock * cp, Eye ** epp,
		  int argc, char ** argv)
{
  SetDefaultValue(cp, argc, argv);
  ReadDataFile(wp, cp, epp);
  ReadCommandParameter(cp, *epp, argc, argv);
}

void InitializeEye(EyeClockWindow * wp, Clock * cp, Eye ** epp)
{
  SetClockValue(cp);
  SetClockGC(wp, cp);
  SetEyeGC(wp, *epp);
  SetEyePixmap(wp, cp, *epp);
}

/*****************************************************************************/
/* ٥Ƚ                                                              */
/*****************************************************************************/

XPoint PointerMotion(EyeClockWindow * wp, Clock * cp, Eye ** epp,
		     int argc, char ** argv, int hour, int minute,
		     XPoint geo, XEvent event)
{
  XPoint pointer;

  pointer.x = event.xmotion.x_root;
  pointer.y = event.xmotion.y_root;
  if (cp->every_flag == ON) {
    FreeGCandEye(wp, epp);
    SetParameter(wp, cp, epp, argc, argv);
    InitializeEye(wp, cp, epp);
  }
  DrawClock(wp, cp, *epp, hour, minute, pointer, geo);
  return (pointer);
}

void ExposeClock(EyeClockWindow * wp, XEvent event)
{
  XCopyArea(wp->display, wp->clock_pixmap, wp->clock_window, wp->gc_copy_copy,
	    event.xexpose.x, event.xexpose.y,
	    event.xexpose.width, event.xexpose.height,
	    event.xexpose.x, event.xexpose.y);
  usleep(1000);
}

void KeyPressClock(EyeClockWindow * wp, Clock * cp, Eye ** epp,
		   int argc, char ** argv, XEvent event)
{
  if (XKeycodeToKeysym(event.xkey.display,
		       event.xkey.keycode, 0) == XK_q) Quit(wp, cp, epp);
  if (XKeycodeToKeysym(event.xkey.display,
		       event.xkey.keycode, 0) == XK_a) {
    cp->animation_flag = ON;
    cp->animation_type = rand() % ANIMATION_NUMBER;
  }
  if (XKeycodeToKeysym(event.xkey.display,
		       event.xkey.keycode, 0) == XK_n) {
    cp->animation_flag = ON;
    cp->animation_type = (cp->animation_type + 1) % ANIMATION_NUMBER;
  }
  if (XKeycodeToKeysym(event.xkey.display,
		       event.xkey.keycode, 0) == XK_w) {
    FreePixmap(wp);
    FreeGCandEye(wp, epp);
    SetParameter(wp, cp, epp, argc, argv);
    CreateClockPixmap(wp, cp);
    InitializeEye(wp, cp, epp);
    SetWindowAttributes(wp, cp);
  }
  if (XKeycodeToKeysym(event.xkey.display,
		       event.xkey.keycode, 0) == XK_r) {
    FreeGCandEye(wp, epp);
    SetParameter(wp, cp, epp, argc, argv);
    InitializeEye(wp, cp, epp);
  }
}

/*****************************************************************************/
/* ̤ܶޤ륢˥᡼                                                */
/*****************************************************************************/

void Animation(EyeClockWindow * wp, Clock * cp, Eye ** epp,
	       int hour, int minute, XPoint geo, int type)
{
  XPoint pointer;
  static int degree = 0;
  
  switch (type) {
  case 0 :
    pointer.x = geo.x + cp->width /2 + 300 * cos(M_PI / 180.0 * degree);
    pointer.y = geo.y + cp->height/2 + 600 * sin(M_PI / 180.0 * degree);
    degree += 45;
    if (degree > 360) degree -= 360;
    break;
  case 1 :
    pointer.x = geo.x + cp->width /2 + 300 * cos(M_PI / 180.0 * degree);
    pointer.y = geo.y + cp->height/2 + 600 * sin(M_PI / 180.0 * degree);
    degree -= 45;
    if (degree > 360) degree += 360;
    break;
  case 2 :
    pointer.x = geo.x + cp->width /2 + 300 * cos(M_PI / 180.0 * degree);
    pointer.y = geo.y + cp->height/2 + 600 * sin(M_PI / 180.0 * (180+degree));
    degree = rand() % 360;
    break;
  }
  DrawClock(wp, cp, *epp, hour, minute, pointer, geo);
}

/*****************************************************************************/
/* ᥤ롼                                                              */
/*****************************************************************************/

void ClockMainLoop(EyeClockWindow * wp, Clock * cp, Eye ** epp,
		   int argc, char ** argv)
{
  int i, hour, minute;
  int configured = True;
  XEvent e, event;
  XPoint pointer, geometry;

  hour = GetHour();
  minute = GetMinute();
  GetGeometry(wp->display, wp->clock_window, &geometry);
  pointer.x = geometry.x + cp->width /2;
  pointer.y = geometry.y + cp->height/2;
  DrawClock(wp, cp, *epp, hour, minute, pointer, geometry);
  while (1) {
    if (minute != GetMinute()) {
      hour = GetHour();
      minute = GetMinute();
      DrawClock(wp, cp, *epp, hour, minute, pointer, geometry);
      if ((minute == 0) || (minute == 30)) {
	cp->animation_flag = ON;
	cp->animation_type = rand() % ANIMATION_NUMBER;
      } else cp->animation_flag = OFF;
    }
    for (i = 0; i < 60; i++) {
      if (cp->eye_flag == ON) {
	if (XCheckTypedWindowEvent(wp->display, wp->toplevel_window,
				   ConfigureNotify, &event) == True)
	  configured = True;
	if (XCheckTypedWindowEvent(wp->display, wp->root_window,
				   MotionNotify, &e) == True) {
	  event = e;
	  while (XCheckTypedWindowEvent(wp->display, wp->root_window,
					MotionNotify, &e) == True) {
	    event = e;
	    usleep(1000);
	  }
	  if (configured == True) {
	    GetGeometry(wp->display, wp->clock_window, &geometry);
	    configured = False;
	  }
	  pointer = PointerMotion(wp, cp, epp, argc, argv, hour, minute,
				  geometry, event);
	  usleep(10000);
	}
      }
      if (XCheckTypedWindowEvent(wp->display, wp->clock_window,
				 Expose, &event) == True)
	ExposeClock(wp, event);
      if (XCheckTypedWindowEvent(wp->display, wp->toplevel_window,
				 ClientMessage, &event) == True) {
	if ( (event.xclient.message_type == wp->atom_1) &&
	     (event.xclient.data.l[0] == wp->atom_2) ) Quit(wp, cp, epp);
      }
      if (XCheckTypedWindowEvent(wp->display, wp->clock_window,
				 KeyPress, &event) == True)
	KeyPressClock(wp, cp, epp, argc, argv, event);
      if (cp->animation_flag == ON)
	Animation(wp, cp, epp, hour, minute, geometry, cp->animation_type);
      usleep(100000);
    }
  }
}

/*****************************************************************************/
/* main ؿ                                                                 */
/*****************************************************************************/

main(int argc, char ** argv)
{
  EyeClockWindow eye_clock_window;
  Clock clock;
  Eye * eyes = NULL;

  srand((unsigned)time(NULL));
  SetParameter(&eye_clock_window, &clock, &eyes, argc, argv);
  if (clock.verbose_flag == ON) CheckData(&clock, eyes);
  CreateWindow(&eye_clock_window, &clock, argc, argv);
  CreateClockPixmap(&eye_clock_window, &clock);
  InitializeEye(&eye_clock_window, &clock, &eyes);
  SetWindowAttributes(&eye_clock_window, &clock);
  ClockMainLoop(&eye_clock_window, &clock, &eyes, argc, argv);
}

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
